from .librrsimulator import libRRSimulator
from .csimulator import CSimulator


def sim_getter(model_file=None, model_str=None, sim_type="libRR"):
    """
    Convenience function to get a simulator object of a specific type.
    Allows you to pull a simulator object given a model file path.

    Note: This likely needs to be refactored but for now it works.

    Parameters
    ----------
    model_file : str, optional
        The path to the model file, at the moment only BNGL is expected
        but this can change in the future.
    model_str : str, optional
        Instead of the path to the model you can also supply the model
        string instead.
    sim_type : str, optional
        The name of the type of simulator object to get. At the moment only
        libRoadRunner and CPY type simulators are allowed, allowed values are
        "libRR" and "cpy".

    Returns
    -------
    BNGSimulator
        A simulator object with an API that's supposed to be agnostic to the
        underlying simulator it's running.
    """
    if model_str is not None and model_file is None:
        from tempfile import NamedTemporaryFile

        with NamedTemporaryFile("w+") as model_file_obj:
            model_file_obj.write(model_str)
            model_file = model_file_obj.name
            if sim_type == "libRR":
                # need to go back to beginning of the file for this to work
                model_file_obj.seek(0)
                return libRRSimulator(model_file=model_file)
            elif sim_type == "cpy":
                return CSimulator(model_file=model_file, generate_network=True)
            else:
                print("simulator type {} not supported".format(sim_type))
    if model_file is not None:
        if sim_type == "libRR":
            return libRRSimulator(model_file=model_file)
        elif sim_type == "cpy":
            return CSimulator(model_file=model_file, generate_network=True)
        else:
            print("simulator type {} not supported".format(sim_type))



import copy, tempfile, shutil

from bionetgen.main import BioNetGen
from bionetgen.core.exc import BNGModelError

from .bngparser import BNGParser
from .blocks import (
    ActionBlock,
    CompartmentBlock,
    FunctionBlock,
    MoleculeTypeBlock,
    ObservableBlock,
    ParameterBlock,
    RuleBlock,
    SpeciesBlock,
    EnergyPatternBlock,
    PopulationMapBlock,
)


# This allows access to the CLIs config setup
app = BioNetGen()
app.setup()
conf = app.config["bionetgen"]
def_bng_path = conf["bngpath"]


###### CORE OBJECT AND PARSING FRONT-END ######
class bngmodel:
    """
    Main model object and entry point for model API. The goal of this
    object is to generate and read the BNGXML of a given BNGL model
    and give the user a pythonic interface to the resulting model object.

    Usage: bngmodel(bng_model)
           bngmodel(bng_model, BNGPATH)

    Attributes
    ----------
    active_blocks : list[str]
        a list of the blocks that have been parsed in the model
    bngparser : BNGParser
        BNGParser object that's responsible for .bngl file reading and model setup
    model_name : str
        name of the model, generally set from the given BNGL file
    model_path : str
        path to the model file initially given
    recompile : bool
        a tag to keep track if any changes have been made to the model
        via the XML API by the user that requires model recompilation
    changes : dict
        a list of changes the user have made to the model

    Methods
    -------
    reset_compilation_tags()
        resets compilation tags of each block to keep track of any changes the user
        makes to the model via the API
    add_action(action_type, action_args)
        adds the action of action_type with arguments given by the optional keyword
        argument action_args, which is a dictionary where each element
        is of the form "ArgumentName":ArgumentValue
    write_model(model_name)
        write the model in BNGL format to the path given
    setup_simulator(sim_type)
        sets up a simulator in bngmodel.simulator where the only current supported
        type of simulator is libRR for libRoadRunner simulator.
    add_block(BlockObject)
        adds a given block object (e.g. ParametersBlock) to the model
    add_empty_block(block_type)
        adds an empty block of type block_type to the model where block_type can be one of: "parameters",
        "compartments", "molecule_types", "species", "observables", "functions", "energy_patterns",
        "population_maps", "rules", "reaction_rules", "actions".
    """

    def __init__(
        self, bngl_model, BNGPATH=def_bng_path, generate_network=False, suppress=True
    ):
        self.active_blocks = []
        # We want blocks to be printed in the same order every time
        self._block_order = [
            "parameters",
            "compartments",
            "molecule_types",
            "species",
            "observables",
            "functions",
            "energy_patterns",
            "population_maps",
            "rules",
            "actions",
        ]
        self.model_name = ""
        self.model_path = bngl_model
        self.bngparser = BNGParser(
            bngl_model, generate_network=generate_network, suppress=True
        )
        self.bngparser.parse_model(self)
        for block in self._block_order:
            if block not in self.active_blocks:
                self.add_empty_block(block)
        # Check to see if there are no active blocks
        # If not, model is most likely not in BNGL format
        if not self.active_blocks:
            # TODO: consider raising a BNGModelError() here
            # raise BNGModelError(
            #                 self.model_path,
            #                 message="WARNING: No active blocks. Please ensure model is in proper BNGL or BNG-XML format",
            #             )
            print(
                "WARNING: No active blocks. Please ensure model is in proper BNGL or BNG-XML format"
            )

    @property
    def recompile(self):
        recompile = False
        for block in self.active_blocks:
            recompile = recompile or getattr(self, block)._recompile
        return recompile

    # TODO: Ensure this works when you edit attributes
    @property
    def changes(self):
        changes = {}
        for block in self.active_blocks:
            changes[block] = getattr(self, block)._changes
        return changes

    def __str__(self):
        """
        write the model to str
        """
        model_str = ""
        # gotta check for "before model" type actions
        if hasattr(self, "actions"):
            ablock = getattr(self, "actions")
            if len(ablock.before_model) > 0:
                for baction in ablock.before_model:
                    model_str += str(baction) + "\n"
        model_str += "begin model\n"
        for block in self._block_order:
            # ensure we didn't get new items into a
            # previously inactive block, if we did
            # add them to the active blocks
            if hasattr(self, block):
                if len(getattr(self, block)) > 0:
                    if getattr(self, block).name not in self.active_blocks:
                        self.active_blocks.append(block)
                # if we removed items from a block and
                # it's now empty, we want to remove it
                # from the active blocks
                elif len(getattr(self, block)) == 0 and block in self.active_blocks:
                    self.active_blocks.remove(block)
            # print only the active blocks
            if block in self.active_blocks:
                if block != "actions" and len(getattr(self, block)) > 0:
                    model_str += str(getattr(self, block))
        model_str += "\nend model\n\n"
        if "actions" in self.active_blocks:
            model_str += str(self.actions)
        return model_str

    def __repr__(self):
        return self.model_name

    def __iter__(self):
        active_ordered_blocks = [
            getattr(self, i) for i in self._block_order if i in self.active_blocks
        ]
        return active_ordered_blocks.__iter__()

    def add_block(self, block):
        """
        Adds the given block object to the model, uses the
        name of the block object to determine what block it is
        """
        bname = block.name.replace(" ", "_")
        # TODO: fix this exception
        if bname == "reaction_rules":
            bname = "rules"
        block_adder = getattr(self, "add_{}_block".format(bname))
        block_adder(block)

    def add_empty_block(self, block_name):
        """
        Makes an empty block object from a given block name and
        adds it to the model object.
        """
        bname = block_name.replace(" ", "_")
        # TODO: fix this exception
        if bname == "reaction_rules":
            bname = "rules"
        block_adder = getattr(self, "add_{}_block".format(bname))
        block_adder()

    def add_parameters_block(self, block=None):
        """
        Adds a parameters block to the model object.
        """
        if block is not None:
            # TODO: Transition to BNGErrors and logging
            assert isinstance(block, ParameterBlock)
            self.parameters = block
            if "parameters" not in self.active_blocks:
                self.active_blocks.append("parameters")
        else:
            self.parameters = ParameterBlock()

    def add_compartments_block(self, block=None):
        """
        Adds a compartments block to the model object.
        """
        if block is not None:
            # TODO: Transition to BNGErrors and logging
            assert isinstance(block, CompartmentBlock)
            self.compartments = block
            if "compartments" not in self.active_blocks:
                self.active_blocks.append("compartments")
        else:
            self.compartments = CompartmentBlock()

    def add_molecule_types_block(self, block=None):
        """
        Adds a molecule types block to the model object.
        """
        if block is not None:
            # TODO: Transition to BNGErrors and logging
            assert isinstance(block, MoleculeTypeBlock)
            self.molecule_types = block
            if "molecule_types" not in self.active_blocks:
                self.active_blocks.append("molecule_types")
        else:
            self.molecule_types = MoleculeTypeBlock()

    def add_species_block(self, block=None):
        """
        Adds a species block to the model object.
        """
        if block is not None:
            # TODO: Transition to BNGErrors and logging
            assert isinstance(block, SpeciesBlock)
            self.species = block
            if "species" not in self.active_blocks:
                self.active_blocks.append("species")
        else:
            self.species = SpeciesBlock()

    def add_observables_block(self, block=None):
        """
        Adds an observable block to the model object.
        """
        if block is not None:
            # TODO: Transition to BNGErrors and logging
            assert isinstance(block, ObservableBlock)
            self.observables = block
            if "observables" not in self.active_blocks:
                self.active_blocks.append("observables")
        else:
            self.observables = ObservableBlock()

    def add_functions_block(self, block=None):
        """
        Adds a functions block to the model object.
        """
        if block is not None:
            # TODO: Transition to BNGErrors and logging
            assert isinstance(block, FunctionBlock)
            self.functions = block
            if "functions" not in self.active_blocks:
                self.active_blocks.append("functions")
        else:
            self.functions = FunctionBlock()

    def add_rules_block(self, block=None):
        """
        Adds a rules block to the model object.
        """
        if block is not None:
            # TODO: Transition to BNGErrors and logging
            assert isinstance(block, RuleBlock)
            self.rules = block
            if "rules" not in self.active_blocks:
                self.active_blocks.append("rules")
        else:
            self.rules = RuleBlock()

    def add_energy_patterns_block(self, block=None):
        """
        Adds an energy patterns block to the model object.
        """
        if block is not None:
            # TODO: Transition to BNGErrors and logging
            assert isinstance(block, EnergyPatternBlock)
            self.energy_patterns = block
            if "energy_patterns" not in self.active_blocks:
                self.active_blocks.append("energy_patterns")
        else:
            self.energy_patterns = EnergyPatternBlock()

    def add_population_maps_block(self, block=None):
        """
        Adds a population maps block to the model object.
        """
        if block is not None:
            # TODO: Transition to BNGErrors and logging
            assert isinstance(block, PopulationMapBlock)
            self.population_maps = block
            if "population_maps" not in self.active_blocks:
                self.active_blocks.append("population_maps")
        else:
            self.population_maps = PopulationMapBlock()

    def add_actions_block(self, block=None):
        """
        Adds an actions block to the model object.
        """
        if block is not None:
            # TODO: Transition to BNGErrors and logging
            assert isinstance(block, ActionBlock)
            self.actions = block
            if "actions" not in self.active_blocks:
                self.active_blocks.append("actions")
        else:
            self.actions = ActionBlock()

    def reset_compilation_tags(self):
        """
        This function resets all internal tags used for keeping
        track of changes done to a model after it's loaded. Resetting
        these tags will remove all history of changes.
        """
        for block in self.active_blocks:
            getattr(self, block).reset_compilation_tags()

    def add_action(self, action_type, action_args={}):
        """
        Adds an action to the actions block of the model object.
        If an actions block doesn't exist, this will make an empty
        actions block and append the action to the block.

        Arguments
        ---------
        action_type: str
            the type of action being added
        action_args: dict
            a dictionary where the key is the argument type and the value
            is the value of that argument.
        """
        # add actions block and to active list
        if not hasattr(self, "actions"):
            self.actions = ActionBlock()
            if "actions" not in self.active_blocks:
                self.active_blocks.append("actions")
        self.actions.add_action(action_type, action_args)

    def write_model(self, file_name):
        """
        write the model to file
        """
        with open(file_name, "w") as f:
            f.write(str(self))

    def setup_simulator(self, sim_type="libRR"):
        """
        Sets up a simulator attribute that is a generic front-end
        to all other simulators. At the moment only libroadrunner
        is supported
        """
        if sim_type == "libRR":
            # we need to add writeSBML action for now
            curr_actions = copy.deepcopy(self.actions)
            self.actions.clear_actions()
            self.add_action("generate_network", {"overwrite": 1})
            self.add_action("writeSBML", {})
            # temporary folder instead to make it work
            # with windows
            try:
                tmp_folder = tempfile.mkdtemp()
                sbml_name = f"{self.model_name}_sbml.xml"
                # write the sbml
                with open(sbml_name, "w+") as f:
                    if not (
                        self.bngparser.bngfile.write_xml(
                            f, xml_type="sbml", bngl_str=str(self)
                        )
                    ):
                        raise BNGModelError(
                            self.model_path,
                            message="SBML couldn't be generated for libRR simulator",
                        )
                self.actions.clear_actions()
                # get the simulator
                import bionetgen as bng

                self.simulator = bng.sim_getter(model_file=sbml_name, sim_type=sim_type)
                # let's deal with observables here
                selections = ["time"] + [obs for obs in self.observables]
                self.simulator.simulator.timeCourseSelections = selections
            finally:
                shutil.rmtree(tmp_folder)
            self.actions = curr_actions
        elif sim_type == "cpy":
            # get the simulator
            import bionetgen as bng

            self.simulator = bng.sim_getter(model_file=self, sim_type=sim_type)
            return self.simulator
        else:
            print(
                'Sim type {} is not recognized, only libroadrunner \
                   is supported currently by passing "libRR" to \
                   sim_type keyword argument'.format(
                    sim_type
                )
            )
            return None
        # for now we return the underlying simulator
        return self.simulator.simulator


###### CORE OBJECT AND PARSING FRONT-END ######



from bionetgen.core.utils.logging import BNGLogger

logger = BNGLogger()


# All classes that deal with patterns
class Pattern:
    """
    Pattern object. Fundamentally it's a list of molecules
    which are defined later.

    Attributes
    ----------
    _bonds : Bonds
        setting a pattern requires you to keep track of all bonds to
        correctly label them, this object tracks everything
    compartment : str
        compartment of the overall pattern (not the same thing as
        molecule compartment, those have their own)
    _label : str
        label of the overall pattern (not the same thing as molecule
        label, those have their own)
    molecules : list[Molecule]
        list of molecule objects that are in the pattern
    fixed : bool
        used for constant species, sets "$" at the beginning of the
        pattern string
    MatchOnce : bool
        used for matchOnce syntax, "{MatchOnce}PatternStr"
    relation : str
        alongside quantity this is only used for patterns of the form
        e.g. pattern==5, pattern<=3 etc
    quantity : str
        alongside relation this is only used for patterns of the form
        e.g. pattern==5, pattern<=3 etc
    nautyG : Graph
        if canonicalization was done on the pattern this will return a graph
        object from the library `pynauty` which is just python bindings to the
        canonical labelling library `nauty`.
    canonical_certificate : str
        if canonicalization was done on the pattern this will return a
        string that has the canonical labelling for the underlying graph.
        this doesn't take into account node names so it can only be used to
        compare the graph topology
    canonical_label : str
        if canonicalization was done on the pattern this will return a canonical
        label for the entire molecule that's unique and all isomorphic patterns will
        have the same label. comparing this label to another molecules canonical label
        will tell you if they are the same molecule or not.

    Methods
    -------
    canonicalize : None
        This method will generate a canonical label stored in `canonical_label` attribute.
        This label can be used to compare patterns to see if they are the same pattern quickly.
        This method will only run if `pynauty` is installed. See [nauty documentation](https://users.cecs.anu.edu.au/~bdm/nauty/)
        for more information
    """

    def __init__(
        self, molecules=[], bonds=None, compartment=None, label=None, canonicalize=False
    ):
        self.molecules = molecules
        self._bonds = bonds
        self.compartment = compartment
        self.label = label
        self.fixed = False
        self.MatchOnce = False
        self.relation = None
        self.quantity = None
        self.nautyG = None
        self.canonical_certificate = None
        self.canonical_label = None
        if canonicalize:
            self.canonicalize()

    def canonicalize(self):
        """
        This method will use `pynauty` library to generate a canonical label
        for the pattern. This pattern will be stored in `canonical_label` attribute.
        """
        # set a location for logging
        loc = f"{__file__} : Pattern.canonicalize()"
        # try importing pynauty to canonicalize the labeling
        try:
            import pynauty
        except ImportError:
            logger.warning(
                f"Importing pynauty failed, cannot canonicalize. Pattern equality checking is not guaranteed to work for highly symmetrical species.",
                loc=loc,
            )
            return
        # find how many vertices we need
        lmol = len(self.molecules)
        lcomp = sum([len(x.components) for x in self.molecules])
        node_cnt = lmol + lcomp
        # initialize our pynauty graph
        G = pynauty.Graph(node_cnt)
        # going to need to figure out bonding
        bond_dict = {}
        # save our IDs
        rev_grpIds = {}
        grpIds = {}
        # also pointers to each object
        node_ptrs = {}
        bond_node_ptrs = {}
        # we'll need to seutp coloring
        colors = {}
        currId = 0
        mCopyId = 0
        cCopyId = 0
        # let's loop over everything in the pattern
        for molec in self.molecules:
            # setting colors
            color_id = (molec.name, None, None)
            if color_id in colors:
                colors[color_id].add(currId)
            else:
                colors[color_id] = set([currId])
            # saving IDs
            parent_id = (molec.name, None, mCopyId, cCopyId)
            if parent_id in grpIds:
                mCopyId += 1
                parent_id = (molec.name, None, mCopyId, cCopyId)
                grpIds[parent_id] = currId
            else:
                grpIds[parent_id] = currId
            rev_grpIds[currId] = parent_id
            node_ptrs[currId] = molec
            currId += 1
            # now looping over components
            for comp in molec.components:
                # saving component coloring
                comp_color_id = (molec.name, comp.name, comp.state)
                if comp_color_id in colors:
                    colors[comp_color_id].add(currId)
                else:
                    colors[comp_color_id] = set([currId])
                chid_id = (molec.name, comp.name, mCopyId, cCopyId)
                # connecting the component to the molecule
                G.connect_vertex(grpIds[parent_id], [currId])
                # saving component IDs
                if chid_id in grpIds:
                    cCopyId += 1
                    chid_id = (molec.name, comp.name, mCopyId, cCopyId)
                    grpIds[chid_id] = currId
                else:
                    grpIds[chid_id] = currId
                rev_grpIds[currId] = chid_id
                node_ptrs[currId] = comp
                currId += 1
                # saving bonds
                if len(comp._bonds) != 0:
                    for bond in comp._bonds:
                        if bond not in bond_dict.keys():
                            bond_dict[bond] = [chid_id]
                        else:
                            bond_dict[bond].append(chid_id)
        # now we got everything, we implement it in the graph
        for bond in bond_dict:
            # check if each of our bonds have exactly two end points
            if len(bond_dict[bond]) == 2:
                id1 = bond_dict[bond][0]
                id1 = grpIds[id1]
                id2 = bond_dict[bond][1]
                id2 = grpIds[id2]
                G.connect_vertex(id1, [id2])
            else:
                # raise a warning
                logger.warning(
                    f"Bond {bond} doesn't have exactly 2 end points, please check that you don't have any dangling bonds.",
                    loc=loc,
                )
        # we get our color sets
        color_sets = list(colors.values())
        # set vertex coloring
        G.set_vertex_coloring(color_sets)
        # save our graph
        self.nautyG = G
        # generate the canonical certificate for the entire graph
        self.canonical_certificate = pynauty.certificate(self.nautyG)
        # generate the canonical label for the entire graph
        # first, we give every node their canonical order
        canon_order = pynauty.canon_label(self.nautyG)
        for iordr, ordr in enumerate(canon_order):
            node_ptrs[ordr].canonical_order = iordr
        # relabeling bonds
        relabeling_bond_dict = {}
        for bond in bond_dict:
            # check if each of our bonds have exactly two end points
            if len(bond_dict[bond]) == 2:
                id1 = bond_dict[bond][0]
                id1 = grpIds[id1]
                comp1 = node_ptrs[id1]
                id2 = bond_dict[bond][1]
                id2 = grpIds[id2]
                comp2 = node_ptrs[id2]
                parent_order = min(
                    comp1.parent_molecule.canonical_order,
                    comp2.parent_molecule.canonical_order,
                )
                comp_order = min(comp1.canonical_order, comp2.canonical_order)
                relabeling_bond_dict[(parent_order, comp_order)] = (comp1, comp2)
            else:
                # raise a warning
                logger.warning(
                    f"Bond {bond} doesn't have exactly 2 end points, please check that you don't have any dangling bonds.",
                    loc=loc,
                )
        # this will give us the keys to canonically sorted bonds
        sorted_order = sorted(relabeling_bond_dict.keys())
        for ibond, sbond in enumerate(sorted_order):
            # now we add a canonical bond ID to each component
            c1, c2 = relabeling_bond_dict[sbond]
            if c1.canonical_bonds is None:
                c1.canonical_bonds = [str(ibond + 1)]
            else:
                c1.canonical_bonds.append(str(ibond + 1))
            if c2.canonical_bonds is None:
                c2.canonical_bonds = [str(ibond + 1)]
            else:
                c2.canonical_bonds.append(str(ibond + 1))
        # and now we can get the canonical label
        self.canonical_label = self.print_canonical()

    def print_canonical(self):
        """
        Returns canonical label for the pattern
        """
        # need to make sure we don't print useless compartments
        self.consolidate_molecule_compartments()
        canon_label = ""
        # we first deal with the pattern compartment
        if self.compartment is not None:
            canon_label += "@{}".format(self.compartment)
        if self.label is not None:
            canon_label += "%{}".format(self.label)
        if self.label is not None or self.compartment is not None:
            canon_label += ":"
        # now loop over all molecules in canonical order
        canon_ords = [m.canonical_order for m in self.molecules]
        canon_ord_pairs = zip(range(len(self.molecules)), canon_ords)
        sorted_canon_ord_pairs = sorted(canon_ord_pairs, key=lambda x: x[1])
        for imol, mol in enumerate(sorted_canon_ord_pairs):
            mol_id = mol[0]
            if imol == 0:
                if self.fixed:
                    canon_label += "$"
                if self.MatchOnce:
                    canon_label += "{MatchOnce}"
            if imol > 0:
                canon_label += "."
            canon_label += self.molecules[mol_id].print_canonical()
        if self.relation is not None:
            canon_label += f"{self.relation}{self.quantity}"
        return canon_label

    def __contains__(self, val):
        return val in self.molecules

    def __eq__(self, other):
        loc = f"{__file__} : Pattern.__eq__()"
        if isinstance(other, Pattern):
            logger.debug(f"Comparison class matches: {other.__class__}", loc=loc)
            # checking pattern-wide properties
            if (other.compartment == self.compartment) and (other.label == self.label):
                logger.debug(
                    f"Compartment or label matches: {other.compartment}, {other.label}",
                    loc=loc,
                )
                # checking mods
                if (other.fixed == self.fixed) and (other.MatchOnce == self.MatchOnce):
                    logger.debug(
                        f"fixed or matchonce matches: {other.fixed}, {other.MatchOnce}",
                        loc=loc,
                    )
                    # checking quantifiers
                    if (other.relation == self.relation) and (
                        other.quantity == self.quantity
                    ):
                        logger.debug(
                            f"relation or quantity matches: {other.relation}, {other.quantity}",
                            loc=loc,
                        )
                        # if we made the label, we can just compare the two
                        if (self.canonical_label is not None) and (
                            other.canonical_label is not None
                        ):
                            return self.canonical_label == other.canonical_label
                        # now we can check contents
                        for molecule in self:
                            if molecule not in other.molecules:
                                logger.debug(
                                    f"molecule doesn't match: {molecule}", loc=loc
                                )
                                return False
                        # isomorphism check if we have the certificate
                        if (self.canonical_certificate is not None) and (
                            other.canonical_certificate is not None
                        ):
                            if (
                                self.canonical_certificate
                                != other.canonical_certificate
                            ):
                                return False
                        # TODO: molecules match, check bonds
                        # Bonds match, patterns are the same
                        logger.debug("patterns match!", loc=loc)
                        return True
        return False

    @property
    def compartment(self):
        return self._compartment

    @compartment.setter
    def compartment(self, value):
        # TODO: Build in logic to set the
        # outer compartment
        # print("Warning: Logical checks are not complete")
        self._compartment = value

    def consolidate_molecule_compartments(self):
        # if the molecule compartment matches overall pattern
        # compartment, don't print the molecule compartments
        overall_comp = self.compartment
        if overall_comp is not None:
            for molec in self.molecules:
                if molec.compartment == overall_comp:
                    molec.compartment = None

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        # TODO: Build in logic to set
        # the outer label
        # print("Warning: Logical checks are not complete")
        self._label = value

    def __str__(self):
        # need to make sure we don't print useless compartments
        self.consolidate_molecule_compartments()
        sstr = ""
        # we first deal with the pattern compartment
        if self.compartment is not None:
            sstr += "@{}".format(self.compartment)
        if self.label is not None:
            sstr += "%{}".format(self.label)
        if self.label is not None or self.compartment is not None:
            sstr += ":"
        # now loop over all molecules
        for imol, mol in enumerate(self.molecules):
            if imol == 0:
                if self.fixed:
                    sstr += "$"
                if self.MatchOnce:
                    sstr += "{MatchOnce}"
            if imol > 0:
                sstr += "."
            sstr += str(mol)
        if self.relation is not None:
            sstr += f"{self.relation}{self.quantity}"
        return sstr

    def __repr__(self):
        return str(self)

    def __getitem__(self, key):
        return self.molecules[key]

    def __iter__(self):
        return self.molecules.__iter__()

    # TODO: Implement __contains__


class Molecule:
    """
    Molecule object. A pattern is a list of molecules.
    This object also handles molecule types where components
    have a list of possible states.

    Attributes
    ----------
    _name : str
        name of the molecule
    _compartment : str
        compartment of the molecule
    _label : str
        label of the molecule
    _components : list[Component]
        list of components for this molecule

    Methods
    -------
    add_component(name, state=None, states=[])
        add a component object to the list of components with name
        "name", current state "state" or a list of states
        (for molecule types) "states"
    """

    def __init__(self, name="0", components=[], compartment=None, label=None):
        self._name = name
        self._components = components
        self._compartment = compartment
        self._label = label
        self.canonical_order = None
        self.canonical_label = None
        self.parent_pattern = None

    def __contains__(self, val):
        return val in self.components

    def __eq__(self, other):
        loc = f"{__file__} : Molecule.__eq__()"
        # check object type
        if isinstance(other, Molecule):
            logger.debug(f"Comparison class matches: {other.__class__}", loc=loc)
            # check attributes
            if (
                (other.name == self.name)
                and (other.compartment == self.compartment)
                and (other.label == self.label)
            ):
                logger.debug(
                    f"name, compartment and labels match: {other.name}, {other.compartment}, {other.label}",
                    loc=loc,
                )
                if (self.canonical_label is not None) and (
                    other.canonical_label is not None
                ):
                    # we can check canonical labels
                    if self.canonical_label != other.canonical_label:
                        return False
                # check components now
                for component in self:
                    if component not in other.components:
                        logger.debug(f"component doesn't match: {component}", loc=loc)
                        return False
                # everything matches
                logger.debug("molecules match", loc=loc)
                return True
        return False

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.components[key]

    def __iter__(self):
        return self.components.__iter__()

    # TODO: implement __setitem__,  __contains__

    def __str__(self):
        mol_str = self.name
        # we have a null species
        if not self.name == "0":
            mol_str += "("
        # we _could_ just not do () if components
        # don't exist but that has other issues,
        # especially for extension highlighting
        if len(self.components) > 0:
            for icomp, comp in enumerate(self.components):
                if icomp > 0:
                    mol_str += ","
                mol_str += str(comp)
        # we have a null species
        if not self.name == "0":
            mol_str += ")"
        if self.compartment is not None:
            mol_str += "@{}".format(self.compartment)
        if self.label is not None:
            mol_str += "%{}".format(self.label)
        return mol_str

    def print_canonical(self):
        """
        Returns canonical label for the pattern
        """
        # print in canonical order
        canon_label = self.name
        # we have a null species
        if not self.name == "0":
            canon_label += "("
        # we _could_ just not do () if components
        # don't exist but that has other issues,
        # especially for extension highlighting
        if len(self.components) > 0:
            canon_ords = [c.canonical_order for c in self.components]
            canon_ord_pairs = zip(range(len(self.components)), canon_ords)
            sorted_canon_ord_pairs = sorted(canon_ord_pairs, key=lambda x: x[1])
            for icomp, comp in enumerate(sorted_canon_ord_pairs):
                comp_id = comp[0]
                if icomp > 0:
                    canon_label += ","
                canon_label += self.components[comp_id].print_canonical()
        # we have a null species
        if not self.name == "0":
            canon_label += ")"
        if self.compartment is not None:
            canon_label += "@{}".format(self.compartment)
        if self.label is not None:
            canon_label += "%{}".format(self.label)
        return canon_label

    ### PROPERTIES ###
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        # print("Warning: Logical checks are not complete")
        # TODO: Check for invalid characters
        self._name = value

    @property
    def components(self):
        return self._components

    @components.setter
    def components(self, value):
        # print("Warning: Logical checks are not complete")
        self._components = value

    def __repr__(self):
        return str(self)

    @property
    def compartment(self):
        return self._compartment

    @compartment.setter
    def compartment(self, value):
        # print("Warning: Logical checks are not complete")
        self._compartment = value

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        # print("Warning: Logical checks are not complete")
        self._label = value

    def _add_component(self, name, state=None, states=[]):
        comp_obj = Component()
        comp_obj.name = name
        comp_obj.state = state
        comp_obj.states = states
        self.components.append(comp_obj)

    def add_component(self, name, state=None, states=[]):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._add_component(name, state, states)


class Component:
    """
    Component object that describes the state, label and bonding
    for each component. Molecules can optionally contain components

    Attributes
    ----------
    name : str
        name of the component
    _label : str
        label of the component
    _state : str
        state of the component, not used for molecule types
    _states : list[str]
        list of states for molecule types
    _bonds : list[Bond]
        list of bond objects that describes bonding of the component

    Methods
    -------
    add_state()
        not implemented. will eventually be used to add additional states
        to an existing component
    add_bond()
        not implemented. will eventually be used to add additional bonds
        to an existing component
    """

    def __init__(self):
        self._name = ""
        self._label = None
        self._state = None
        self._states = []
        self._bonds = []
        self.canonical_label = None
        self.canonical_order = None
        self.canonical_bonds = None
        self.parent_molecule = None

    def __eq__(self, other):
        loc = f"{__file__} : Component.__eq__()"
        # check type
        # import ipdb;ipdb.set_trace()
        if isinstance(other, Component):
            logger.debug(f"Comparison class matches: {other.__class__}", loc=loc)
            # check attributes
            if (other.name == self.name) and (other.label == self.label):
                logger.debug(
                    f"name and labels match: {other.name}, {other.label}", loc=loc
                )
                # check states
                if len(other.states) == len(self.states):
                    logger.debug(f"state lists match: {other.states}", loc=loc)
                    # check current state
                    if other.state == self.state:
                        logger.debug(f"states match: {other.state}", loc=loc)
                        if (self.canonical_label is not None) and (
                            other.canonical_label is not None
                        ):
                            # we can check canonical labels
                            if self.canonical_label != other.canonical_label:
                                return False
                        # check bonds
                        # TODO: try to decide if A(b!1).B(a!1) is the same
                        # as A(b!2).B(a!2), if so, the bond check is much harder
                        # for bond in self.bonds:
                        #     if bond not in other.bonds:
                        #         logger.debug(
                        #             f"bonds don't match!: {other.bonds}", loc=loc
                        #         )
                        #         return False
                        if len(self.bonds) == len(other.bonds):
                            logger.debug("components match", loc=loc)
                            return True
        return False

    def __repr__(self):
        return str(self)

    def __str__(self):
        comp_str = self.name
        # only for molecule types
        if len(self.states) > 0:
            for istate, state in enumerate(self.states):
                comp_str += "~{}".format(state)
        # for any other pattern
        if self.state is not None:
            comp_str += "~{}".format(self.state)
        if self.label is not None:
            comp_str += "%{}".format(self.label)
        if len(self.bonds) > 0:
            for bond in self.bonds:
                comp_str += "!{}".format(bond)
        return comp_str

    def print_canonical(self):
        """
        Returns canonical label for the pattern
        """
        comp_str = self.name
        # only for molecule types
        if len(self.states) > 0:
            for istate, state in enumerate(self.states):
                comp_str += "~{}".format(state)
        # for any other pattern
        if self.state is not None:
            comp_str += "~{}".format(self.state)
        if self.label is not None:
            comp_str += "%{}".format(self.label)
        if self.canonical_bonds is not None:
            for bond in self.canonical_bonds:
                comp_str += "!{}".format(bond)
        return comp_str

    ### PROPERTIES ###
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._name = value

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._label = value

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._state = value

    @property
    def states(self):
        return self._states

    @states.setter
    def states(self, value):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._states = value

    @property
    def bonds(self):
        return self._bonds

    @bonds.setter
    def bonds(self, value):
        # TODO: Add built-in logic here
        # print("Warning: Logical checks are not complete")
        self._bonds = value

    def _add_state(self):
        raise NotImplementedError

    def add_state(self):
        self._add_state()

    def _add_bond(self):
        raise NotImplementedError

    def add_bond(self):
        self._add_bond()



import os, glob
from pytest import raises
import bionetgen as bng
from bionetgen.main import BioNetGenTest

tfold = os.path.dirname(__file__)


def test_bionetgen_help():
    # tests basic command help
    with raises(SystemExit):
        argv = ["--help"]
        with BioNetGenTest(argv=argv) as app:
            app.run()
            assert app.exit_code == 0


def test_bionetgen_input():
    argv = [
        "run",
        "-i",
        os.path.join(tfold, "test.bngl"),
        "-o",
        os.path.join(tfold, "test"),
    ]
    to_match = ["test.xml", "test.cdat", "test.gdat", "test.net"]
    with BioNetGenTest(argv=argv) as app:
        app.run()
        assert app.exit_code == 0
        file_list = os.listdir(os.path.join(tfold, "test"))
        assert file_list.sort() == to_match.sort()


def test_bionetgen_plot():
    argv = [
        "plot",
        "-i",
        os.path.join(*[tfold, "test", "test.gdat"]),
        "-o",
        os.path.join(*[tfold, "test", "test.png"]),
    ]
    with BioNetGenTest(argv=argv) as app:
        app.run()
        assert app.exit_code == 0
        assert os.path.isfile(os.path.join(*[tfold, "test", "test.png"]))


def test_bionetgen_info():
    # tests info subcommand
    argv = ["info"]
    with BioNetGenTest(argv=argv) as app:
        app.run()
        assert app.exit_code == 0



import time, json, urllib, sys
import urllib.request

# GH API has rate exceeded issues if more than 60
# within an hour. To avoid this entirely, we'll
# download and save the JSON and have setup.py
# work on that instead

# let's pull URLs for each distribution
# in the latest distribution
rls_url = "https://api.github.com/repos/RuleWorld/bionetgen/releases/latest"
# sometimes we exceed the rate, we want to
# ensure this doesn't happen, but we also
# _must_ get the response. We'll slap it into
# a loop and break if need be
ctr = 0
while ctr < 100:
    ctr += 1
    try:
        time.sleep(5)
        rls_resp = urllib.request.urlopen(rls_url)
        print(f"success: {ctr}")
        break
    except urllib.error.HTTPError:
        time.sleep(5)
        print(f"failed: {ctr}")
if ctr >= 100:
    print("Connection to GitHub couldn't be established, quitting")
    sys.exit(1)
rls_json_txt = rls_resp.read()
rls_json = json.loads(rls_json_txt)

with open("ghapi.json", "w") as f:
    json.dump(rls_json, f)



import os, glob
from pytest import raises
import bionetgen as bng
from bionetgen.main import BioNetGenTest

tfold = os.path.dirname(__file__)


def test_bionetgen_model():
    fpath = os.path.join(tfold, "test.bngl")
    fpath = os.path.abspath(fpath)
    m = bng.bngmodel(fpath)


def test_bionetgen_all_model_loading():
    # tests library model loading using many models
    mpattern = os.path.join(tfold, "models") + os.sep + "*.bngl"
    models = glob.glob(mpattern)
    succ = []
    fail = []
    success = 0
    fails = 0
    for model in models:
        try:
            m = bng.bngmodel(model)
            success += 1
            mstr = str(m)
            succ.append(model)
        except:
            print("can't load model {}".format(model))
            fails += 1
            fail.append(model)
    print("succ: {}".format(success))
    print(sorted(succ))
    print("fail: {}".format(fails))
    print(sorted(fail))
    assert fails == 0


def test_action_loading():
    # tests a BNGL file containing all BNG actions
    all_action_model = os.path.join(*[tfold, "models", "actions", "all_actions.bngl"])
    m1 = bng.bngmodel(all_action_model)
    assert len(m1.actions) + len(m1.actions.before_model) == 31

    no_action_model = os.path.join(*[tfold, "models", "actions", "no_actions.bngl"])
    m2 = bng.bngmodel(no_action_model)
    assert len(m2.actions) == 0


def test_model_running_CLI():
    # tests running a list of models using the CLI
    mpattern = os.path.join(tfold, "models") + os.sep + "*.bngl"
    models = glob.glob(mpattern)
    succ = []
    fail = []
    success = 0
    fails = 0
    test_run_folder = os.path.join(tfold, "models", "cli_test_runs")
    if not os.path.isdir(test_run_folder):
        os.mkdir(test_run_folder)
    for model in models:
        model_name = os.path.basename(model).replace(".bngl", "")
        try:
            argv = [
                "run",
                "-i",
                model,
                "-o",
                os.path.join(*[tfold, "models", "cli_test_runs", model_name]),
            ]
            with BioNetGenTest(argv=argv) as app:
                app.run()
                assert app.exit_code == 0
            success += 1
            model = os.path.split(model)
            model = model[1]
            succ.append(model)
        except:
            print("can't run model {}".format(model))
            fails += 1
            model = os.path.split(model)
            model = model[1]
            fail.append(model)
    print("succ: {}".format(success))
    print(sorted(succ))
    print("fail: {}".format(fails))
    print(sorted(fail))
    assert fails == 0


def test_model_running_lib():
    # test running a list of models using the library
    mpattern = os.path.join(tfold, "models") + os.sep + "*.bngl"
    models = glob.glob(mpattern)
    succ = []
    fail = []
    success = 0
    fails = 0
    for model in models:
        if "test_tfun" in model:
            continue
        try:
            bng.run(model)
            success += 1
            model = os.path.split(model)
            model = model[1]
            succ.append(model)
        except:
            print("can't run model {}".format(model))
            fails += 1
            model = os.path.split(model)
            model = model[1]
            fail.append(model)
    print("succ: {}".format(success))
    print(sorted(succ))
    print("fail: {}".format(fails))
    print(sorted(fail))
    assert fails == 0


def test_setup_simulator():
    fpath = os.path.join(tfold, "test.bngl")
    fpath = os.path.abspath(fpath)
    try:
        m = bng.bngmodel(fpath)
        librr_simulator = m.setup_simulator()
        res = librr_simulator.simulate(0, 1, 10)
    except:
        res = None
    assert res is not None



import os, glob
from pytest import raises
import bionetgen as bng
from bionetgen.main import BioNetGenTest

tfold = os.path.dirname(__file__)


def test_bionetgen_visualize():
    vis_types = [
        "contactmap",
        "ruleviz_pattern",
        "ruleviz_operation",
        "regulatory",
        "atom_rule",
        "all",
    ]
    for vis_name in vis_types:
        argv = [
            "visualize",
            "-i",
            os.path.join(tfold, "test.bngl"),
            "-o",
            os.path.join(tfold, "viz"),
            "-t",
            vis_name,
        ]
        with BioNetGenTest(argv=argv) as app:
            app.run()
            assert app.exit_code == 0
            # gmls = glob.glob("*.gml")
            graphmls = glob.glob(os.path.join(tfold, "viz") + os.sep + "*.graphml")
            if vis_name == "atom_rule":
                assert any(["regulatory" in i for i in graphmls])
            elif not vis_name == "all":
                assert any([vis_name in i for i in graphmls])
            else:
                assert len(graphmls) == 4


# def test_graphdiff_matrix():
#     valid = []
#     invalid = []
#     argv = [
#         "graphdiff",
#         "-i",
#         os.path.join(*[tfold, "models", "testviz1_cm.graphml"]),
#         "-i2",
#         os.path.join(*[tfold, "models", "testviz2_cm.graphml"]),
#         "-m",
#         "matrix",
#     ]
#     to_validate = ["testviz1_cm_recolored.graphml",
#                 "testviz1_cm_testviz2_cm_diff.graphml",
#                 "testviz2_cm_recolored.graphml",
#                 "testviz2_cm_testviz1_cm_diff.graphml",
#                 ]
#     schema_doc = etree.parse(f)
#     xmlschema = etree.XMLSchema(schema_doc)

#     with BioNetGenTest(argv=argv) as app:
#         app.run()
#         assert app.exit_code == 0
#     for test_graphml in to_validate:
#         doc = etree.parse(test_graphml)
#         result = xmlschema.validate(doc)
#         if result == True: valid.append(test_graphml)
#         else:
#             invalid.append(test_graphml)
#     print(sorted(valid))
#     print(sorted(invalid))
#     # assert len(valid) == 4


# def test_graphdiff_union():
#     argv = [
#         "graphdiff",
#         "-i",
#         os.path.join(tfold, "models", "testviz1_cm.graphml"),
#         "-i2",
#         os.path.join(tfold, "models", "testviz2_cm.graphml"),
#         "-m",
#         "union",
#     ]
#     to_validate = "testviz1_cm_testviz2_cm_union.graphml"
#     # xmlschema_doc = etree.parse("INSERT_xsd_path_HERE.xsd")
#     # xmlschema = etree.XMLSchema(xmlschema_doc)
#     with BioNetGenTest(argv=argv) as app:
#         app.run()
#         assert app.exit_code == 0
#     # xml_doc = etree.parse(to_validate)
#     # result = xmlschema.validate(xml_doc)
#     # assert result == True



import os, glob
from pytest import raises
import bionetgen as bng
from bionetgen.main import BioNetGenTest

tfold = os.path.dirname(__file__)


def test_bionetgen_help():
    # tests basic command help
    with raises(SystemExit):
        argv = ["--help"]
        with BioNetGenTest(argv=argv) as app:
            app.run()
            assert app.exit_code == 0


def test_bionetgen_input():
    argv = [
        "run",
        "-i",
        os.path.join(tfold, "test.bngl"),
        "-o",
        os.path.join(tfold, "test"),
    ]
    to_match = ["test.xml", "test.cdat", "test.gdat", "test.net"]
    with BioNetGenTest(argv=argv) as app:
        app.run()
        assert app.exit_code == 0
        file_list = os.listdir(os.path.join(tfold, "test"))
        assert file_list.sort() == to_match.sort()


def test_bionetgen_plot():
    argv = [
        "plot",
        "-i",
        os.path.join(*[tfold, "test", "test.gdat"]),
        "-o",
        os.path.join(*[tfold, "test", "test.png"]),
    ]
    with BioNetGenTest(argv=argv) as app:
        app.run()
        assert app.exit_code == 0
        assert os.path.isfile(os.path.join(*[tfold, "test", "test.png"]))


def test_bionetgen_model():
    fpath = os.path.join(tfold, "test.bngl")
    fpath = os.path.abspath(fpath)
    m = bng.bngmodel(fpath)


def test_bionetgen_visualize():
    vis_types = [
        "contactmap",
        "ruleviz_pattern",
        "ruleviz_operation",
        "regulatory",
        "atom_rule",
        "all",
    ]
    for vis_name in vis_types:
        argv = [
            "visualize",
            "-i",
            os.path.join(tfold, "test.bngl"),
            "-o",
            os.path.join(tfold, "viz"),
            "-t",
            vis_name,
        ]
        with BioNetGenTest(argv=argv) as app:
            app.run()
            assert app.exit_code == 0
            # gmls = glob.glob("*.gml")
            graphmls = glob.glob(os.path.join(tfold, "viz") + os.sep + "*.graphml")
            if vis_name == "atom_rule":
                assert any(["regulatory" in i for i in graphmls])
            elif not vis_name == "all":
                assert any([vis_name in i for i in graphmls])
            else:
                assert len(graphmls) == 4


def test_bionetgen_all_model_loading():
    # tests library model loading using many models
    mpattern = os.path.join(tfold, "models") + os.sep + "*.bngl"
    models = glob.glob(mpattern)
    succ = []
    fail = []
    success = 0
    fails = 0
    for model in models:
        try:
            m = bng.bngmodel(model)
            success += 1
            mstr = str(m)
            succ.append(model)
        except:
            print("can't load model {}".format(model))
            fails += 1
            fail.append(model)
    print("succ: {}".format(success))
    print(sorted(succ))
    print("fail: {}".format(fails))
    print(sorted(fail))
    assert fails == 0


def test_action_loading():
    # tests a BNGL file containing all BNG actions
    all_action_model = os.path.join(*[tfold, "models", "actions", "all_actions.bngl"])
    m1 = bng.bngmodel(all_action_model)
    assert len(m1.actions) + len(m1.actions.before_model) == 31

    no_action_model = os.path.join(*[tfold, "models", "actions", "no_actions.bngl"])
    m2 = bng.bngmodel(no_action_model)
    assert len(m2.actions) == 0


def test_bionetgen_info():
    # tests info subcommand
    argv = ["info"]
    with BioNetGenTest(argv=argv) as app:
        app.run()
        assert app.exit_code == 0


def test_model_running_CLI():
    # tests running a list of models using the CLI
    mpattern = os.path.join(tfold, "models") + os.sep + "*.bngl"
    models = glob.glob(mpattern)
    succ = []
    fail = []
    success = 0
    fails = 0
    test_run_folder = os.path.join(tfold, "models", "cli_test_runs")
    if not os.path.isdir(test_run_folder):
        os.mkdir(test_run_folder)
    for model in models:
        model_name = os.path.basename(model).replace(".bngl", "")
        try:
            argv = [
                "run",
                "-i",
                model,
                "-o",
                os.path.join(*[tfold, "models", "cli_test_runs", model_name]),
            ]
            with BioNetGenTest(argv=argv) as app:
                app.run()
                assert app.exit_code == 0
            success += 1
            model = os.path.split(model)
            model = model[1]
            succ.append(model)
        except:
            print("can't run model {}".format(model))
            fails += 1
            model = os.path.split(model)
            model = model[1]
            fail.append(model)
    print("succ: {}".format(success))
    print(sorted(succ))
    print("fail: {}".format(fails))
    print(sorted(fail))
    assert fails == 0


def test_model_running_lib():
    # test running a list of models using the library
    mpattern = os.path.join(tfold, "models") + os.sep + "*.bngl"
    models = glob.glob(mpattern)
    succ = []
    fail = []
    success = 0
    fails = 0
    for model in models:
        if "test_tfun" in model:
            continue
        try:
            bng.run(model)
            success += 1
            model = os.path.split(model)
            model = model[1]
            succ.append(model)
        except:
            print("can't run model {}".format(model))
            fails += 1
            model = os.path.split(model)
            model = model[1]
            fail.append(model)
    print("succ: {}".format(success))
    print(sorted(succ))
    print("fail: {}".format(fails))
    print(sorted(fail))
    assert fails == 0


def test_atomize_flat():
    if not os.path.exists(os.path.join(tfold, "test")):
        os.mkdir(os.path.join(tfold, "test"))
    argv = [
        "atomize",
        "-i",
        os.path.join(tfold, "test_sbml.xml"),
        "-o",
        os.path.join(*[tfold, "test", "test_sbml_flat.bngl"]),
    ]
    to_match = ["test_sbml_flat.bngl"]
    with BioNetGenTest(argv=argv) as app:
        app.run()
        assert app.exit_code == 0
        file_list = os.listdir(os.path.join(tfold, "test"))
        assert file_list.sort() == to_match.sort()


def test_atomize_atomized():
    if not os.path.exists(os.path.join(tfold, "test")):
        os.mkdir(os.path.join(tfold, "test"))
    argv = [
        "atomize",
        "-i",
        os.path.join(tfold, "test_sbml.xml"),
        "-o",
        os.path.join(*[tfold, "test", "test_sbml_atom.bngl"]),
        "-a",
    ]
    to_match = ["test_sbml_atom.bngl"]
    with BioNetGenTest(argv=argv) as app:
        app.run()
        assert app.exit_code == 0
        file_list = os.listdir(os.path.join(tfold, "test"))
        assert file_list.sort() == to_match.sort()


def test_network_parse():
    netfile = os.path.join(tfold, "mockup.net")
    from bionetgen.network.network import Network

    try:
        net = Network(netfile)
        res = True
    except:
        res = False
    assert res is True


def test_pattern_reader():
    patfile = os.path.join(tfold, "patterns.txt")
    from bionetgen.modelapi.pattern_reader import BNGPatternReader

    try:
        with open(patfile, "r") as f:
            patterns = f.readlines()
            for pattern in patterns:
                pat_obj = BNGPatternReader(pattern).pattern
                reparsed_pat = BNGPatternReader(str(pat_obj)).pattern
                if pat_obj != reparsed_pat:
                    raise RuntimeError(
                        f"Pattern can't be reparsed correctly, og: {pat_obj}, reparsed: {reparsed_pat}"
                    )
        res = True
    except:
        res = False
    assert res is True


def test_pattern_canonicalization():
    # for now, if the platform is windows, just skip
    if os.name == "nt":
        assert True is True
    # if pynauty is uninstalled, skip the test
    try:
        import pynauty
    except ImportError:
        assert True is True
        return
    # otherwise we will test canonicalization
    from bionetgen.modelapi.pattern_reader import BNGPatternReader

    # the testing file
    testfile = os.path.join(tfold, "canon_label_testing.txt")
    with open(testfile, "r+") as f:
        tests = f.readlines()
    # loop over tests
    res = True
    for ipat, pat in enumerate(tests):
        pat_splt = pat.split("    ")
        pat1, pat2 = pat_splt[0], pat_splt[1]
        try:
            # read patterns
            pat1_obj = BNGPatternReader(pat1).pattern
            pat2_obj = BNGPatternReader(pat2).pattern
            # compare them
            if pat1_obj != pat2_obj:
                res = False
                break
        except:
            res = False
            break
    # assert that everything matched up
    assert res is True


def test_setup_simulator():
    fpath = os.path.join(tfold, "test.bngl")
    fpath = os.path.abspath(fpath)
    try:
        m = bng.bngmodel(fpath)
        librr_simulator = m.setup_simulator()
        res = librr_simulator.simulate(0, 1, 10)
    except:
        res = None
    assert res is not None


# def test_graphdiff_matrix():
#     valid = []
#     invalid = []
#     argv = [
#         "graphdiff",
#         "-i",
#         os.path.join(*[tfold, "models", "testviz1_cm.graphml"]),
#         "-i2",
#         os.path.join(*[tfold, "models", "testviz2_cm.graphml"]),
#         "-m",
#         "matrix",
#     ]
#     to_validate = ["testviz1_cm_recolored.graphml",
#                 "testviz1_cm_testviz2_cm_diff.graphml",
#                 "testviz2_cm_recolored.graphml",
#                 "testviz2_cm_testviz1_cm_diff.graphml",
#                 ]
#     schema_doc = etree.parse(f)
#     xmlschema = etree.XMLSchema(schema_doc)

#     with BioNetGenTest(argv=argv) as app:
#         app.run()
#         assert app.exit_code == 0
#     for test_graphml in to_validate:
#         doc = etree.parse(test_graphml)
#         result = xmlschema.validate(doc)
#         if result == True: valid.append(test_graphml)
#         else:
#             invalid.append(test_graphml)
#     print(sorted(valid))
#     print(sorted(invalid))
#     # assert len(valid) == 4


# def test_graphdiff_union():
#     argv = [
#         "graphdiff",
#         "-i",
#         os.path.join(tfold, "models", "testviz1_cm.graphml"),
#         "-i2",
#         os.path.join(tfold, "models", "testviz2_cm.graphml"),
#         "-m",
#         "union",
#     ]
#     to_validate = "testviz1_cm_testviz2_cm_union.graphml"
#     # xmlschema_doc = etree.parse("INSERT_xsd_path_HERE.xsd")
#     # xmlschema = etree.XMLSchema(xmlschema_doc)
#     with BioNetGenTest(argv=argv) as app:
#         app.run()
#         assert app.exit_code == 0
#     # xml_doc = etree.parse(to_validate)
#     # result = xmlschema.validate(xml_doc)
#     # assert result == True



import os, glob
from pytest import raises
import bionetgen as bng
from bionetgen.main import BioNetGenTest

tfold = os.path.dirname(__file__)


def test_network_parse():
    netfile = os.path.join(tfold, "mockup.net")
    from bionetgen.network.network import Network

    try:
        net = Network(netfile)
        res = True
    except:
        res = False
    assert res is True


def test_pattern_reader():
    patfile = os.path.join(tfold, "patterns.txt")
    from bionetgen.modelapi.pattern_reader import BNGPatternReader

    try:
        with open(patfile, "r") as f:
            patterns = f.readlines()
            for pattern in patterns:
                pat_obj = BNGPatternReader(pattern).pattern
                reparsed_pat = BNGPatternReader(str(pat_obj)).pattern
                if pat_obj != reparsed_pat:
                    raise RuntimeError(
                        f"Pattern can't be reparsed correctly, og: {pat_obj}, reparsed: {reparsed_pat}"
                    )
        res = True
    except:
        res = False
    assert res is True


def test_pattern_canonicalization():
    # for now, if the platform is windows, just skip
    if os.name == "nt":
        assert True is True
    # if pynauty is uninstalled, skip the test
    try:
        import pynauty
    except ImportError:
        assert True is True
        return
    # otherwise we will test canonicalization
    from bionetgen.modelapi.pattern_reader import BNGPatternReader

    # the testing file
    testfile = os.path.join(tfold, "canon_label_testing.txt")
    with open(testfile, "r+") as f:
        tests = f.readlines()
    # loop over tests
    res = True
    for ipat, pat in enumerate(tests):
        pat_splt = pat.split("    ")
        pat1, pat2 = pat_splt[0], pat_splt[1]
        try:
            # read patterns
            pat1_obj = BNGPatternReader(pat1).pattern
            pat2_obj = BNGPatternReader(pat2).pattern
            # compare them
            if pat1_obj != pat2_obj:
                res = False
                break
        except:
            res = False
            break
    # assert that everything matched up
    assert res is True



import os, glob
from pytest import raises
import bionetgen as bng
from bionetgen.main import BioNetGenTest

tfold = os.path.dirname(__file__)


def test_atomize_flat():
    if not os.path.exists(os.path.join(tfold, "test")):
        os.mkdir(os.path.join(tfold, "test"))
    argv = [
        "atomize",
        "-i",
        os.path.join(tfold, "test_sbml.xml"),
        "-o",
        os.path.join(*[tfold, "test", "test_sbml_flat.bngl"]),
    ]
    to_match = ["test_sbml_flat.bngl"]
    with BioNetGenTest(argv=argv) as app:
        app.run()
        assert app.exit_code == 0
        file_list = os.listdir(os.path.join(tfold, "test"))
        assert file_list.sort() == to_match.sort()


def test_atomize_atomized():
    if not os.path.exists(os.path.join(tfold, "test")):
        os.mkdir(os.path.join(tfold, "test"))
    argv = [
        "atomize",
        "-i",
        os.path.join(tfold, "test_sbml.xml"),
        "-o",
        os.path.join(*[tfold, "test", "test_sbml_atom.bngl"]),
        "-a",
    ]
    to_match = ["test_sbml_atom.bngl"]
    with BioNetGenTest(argv=argv) as app:
        app.run()
        assert app.exit_code == 0
        file_list = os.listdir(os.path.join(tfold, "test"))
        assert file_list.sort() == to_match.sort()
